// cfpfactory.C

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{


/*
Basic factory operation is a little convoluted:
1) a global factory (GlobalFactory) is created on program startup automatically
2) each cfplib library is represented by one or more CfpObjects which are
    spawners, responsible for telling the library about themselves.
3) to do this automatically, each of these objects should be instantiated once using
    a global static declaration, using a no-argument constructor.
4) this no-argument constructor needs only call the base no-argument constructor
    for the CfpObject class.
5) this base class constructor will register itself with the factory as a spawning object.
6) this will be done for all configpp libs prior to execution of the main program,
    due to global instantiations.
7) When the program starts, the factory will be given a pointer to a configpp class, and commandline args
8) The factory will then inform each of its registered spawning objects about
    the command line args, give them a pointer to a configpp class,
    and ask them to register any classes, which is the whole purpose of all this.
*/





//--------------------------------------------------------------------------------
//---- CfpFactory::CfpFactory() --------------------------------------------------
//--------------------------------------------------------------------------------

CfpFactory::CfpFactory()
{
  // constructor;
  configppp=NULL;
  newspawners=false;

  // create the dummy spawner class which spawners may point to
  spawnerclassp=new CfpClass(this,"FactorySpawenerClass","dummyfactorypackage",NULL);
}


//--------------------------------------------------------------------------------
//---- CfpFactory::~CfpFactory() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpFactory::~CfpFactory()
{
  // destructor;
  if (configppp!=NULL)
    configppp->debug("Destructing the class factory.");

  // we should delete all registered classes
  CfpClass *classp=NULL;
  map<string, CfpClass *>::iterator p=classmanager.begin();

  // delete the classes
  while (p!=classmanager.end())
    {
      classp=p->second;
      delete classp;
      p++;
    }

  // don't delete the spawners since they are most likely global and static
}



void CfpFactory::setconfigppp(Configpp *inconfigppp)
{
  // we have a pointer to a configpp object through which we can communicate
  // NOTE that in the future we are almost set up to let one factory manager several instances
  //  of configpp interpreter.  the only thing we would have to do is keep track of multiple configppp pointers.
  //  but for now, just assume one.

  // set the configpp pointer
  configppp=inconfigppp;
}



void CfpFactory::updatespawners()
{
  // Time to tell any new spawners that they need to register with us
  CfpClass *classp;
  CfpObject *objp;
  deque<CfpObject *>::iterator op;

  // first make sure we want to do this now
  if (!newspawners || !configppp)
    return;

  do
    {
      // clear the newspawners flag
      // note we do this before the loop, just in case someone tries to recursively add a spawner while we are registering them.
      newspawners=false;
      
      // now go through the list and make them register
      op=pending.begin();
      while (op!=pending.end())
	{
	  // tell it to register
	  objp=*op;
	  //	  cout << "... asking "+objp->prettyname()+" to register its classes.\n";
	  objp->setconfigppp(configppp);
	  objp->registerwrappers(this);
	  // and now move the spawner to the list of completely registered spawners
	  pending.pop_front();
	  spawnermanager.push_back(objp);
	  op++;
	}
    } while (newspawners!=false);
}



void CfpFactory::registerspawner(CfpObject *spawnerp,string packagename)
{
  // a spawner object is registering with us

  if (configppp!=NULL)
    {
      // let's just do it now
      spawnerp->setconfigppp(configppp);
      spawnerp->registerwrappers(this);
    }
  else
    {
      // add the spawner to our list, and mark it as having NOT been told to register its classes with us yet.
      pending.push_back(spawnerp);
      // set flag saying that we have spawners pending that need to eventually register their classes with us
      newspawners=true;
    }
}



void CfpFactory::viewcommandline(int &argc,char **argv)
{
  // let all our registered spawners take a crack at the commandline (this should be called after setconfigppp above).
  deque<CfpObject *>::iterator p;

  // before any operations on the spawnermanager list, we need to update it to make sure the latest spawners have registered
  updatespawners();

  p=spawnermanager.begin();
  while (p!=spawnermanager.end())
    {
      (*p)->viewcommandline(argc,argv);
      p++;
    }
}




//--------------------------------------------------------------------------------
//---- CfpFactory::registerclass() -----------------------------------------------
//--------------------------------------------------------------------------------

CfpClass *CfpFactory::registerclass(string packagename, string classname, CfpObject *spawnerp)
{
  // register a new library class
  CfpClass *classp;

  // make a new class entry
  classp=new CfpClass(this,classname,packagename,spawnerp);

  //  cout << "... Registering class "+packagename+":"+classname+"\n";
  configppp->debug("registered class "+classp->getname()+" from package " +classp->getpackagename()+".");

  // add the class to the classmanager map
  classmanager.insert(make_pair(classname,classp));

  // lastly, return a pointer to the class that was constructed
  return classp;
}




//--------------------------------------------------------------------------------
//---- CfpFactory::lookupClass() -------------------------------------------------
//--------------------------------------------------------------------------------

CfpClass *CfpFactory::lookupClass(string classname, bool throwerr)
{
  // smart lookup a library class, and return a pointer to it.
  // if throwerr is true, then throw an error if we can't locat the class (otherwise just return NULL)
  CfpClass *classp=NULL;
  map<string, CfpClass *>::iterator p;

  // we need to make sure the latest spawners have registered
  updatespawners();

  p=classmanager.find(classname);
  if (p!=classmanager.end())
    classp=p->second;
  else
    {
      // we couldn't find the class
      if (throwerr)
	throw CfpError("object class '" +classname+ "' is not known.","",classerr,false,false);
    }
  return classp;
}



//--------------------------------------------------------------------------------
//---- CfpFactory::registerunit() ------------------------------------------------
//--------------------------------------------------------------------------------

CfpUnit *CfpFactory::registerunit(CfpUnit *newunitp)
{
  // register a new unit object
  string uname=newunitp->getname();
  
  if (uname!="" && configppp!=NULL)
    configppp->debug("registering unit measurement '" +uname+ "'.");

  // add the unit object to the unit manager
  unitmanager.insert(make_pair(uname,newunitp));
  return newunitp;
}



//--------------------------------------------------------------------------------
//---- CfpFactory::registertype()
//--------------------------------------------------------------------------------

CfpType *CfpFactory::registertype(CfpType *newtypep)
{
  // register a new type object
  string tname=newtypep->getname();
  
  if (tname!="" && configppp!=NULL)
    configppp->debug("registering type '" +tname+ "'.");

  // add the type object to the unit manager
  typemanager.insert(make_pair(tname,newtypep));
  return newtypep;
}




//--------------------------------------------------------------------------------
//---- CfpFactory::lookupUnit() --------------------------------------------------
//--------------------------------------------------------------------------------

CfpUnit *CfpFactory::lookupUnit(string unitname, bool throwerr)
{
  // smart lookup a unit object, and return a pointer to it.
  // if throwerr is true, then throw an error if we can't locate it (otherwise just return NULL)
  CfpUnit *unitp=NULL;
  map<string, CfpUnit *>::iterator p;

  p=unitmanager.find(unitname);
  if (p!=unitmanager.end())
    unitp=p->second;
  else
    {
      // we couldn't find it
      if (throwerr)
	throw CfpError("unit measurement '" +unitname+ "' is not known.","",classerr,false,false);
    }
  return unitp;
}



//--------------------------------------------------------------------------------
//---- CfpFactory::lookupType()
//--------------------------------------------------------------------------------

CfpType *CfpFactory::lookupType(string typname, bool throwerr)
{
  // smart lookup a type object, and return a pointer to it.
  // if throwerr is true, then throw an error if we can't locate it (otherwise just return NULL)
  CfpType *typep=NULL;
  map<string, CfpType *>::iterator p;

  p=typemanager.find(typname);
  if (p!=typemanager.end())
    typep=p->second;
  else
    {
      // we couldn't find it
      if (throwerr)
	throw CfpError("variable/argument type '" +typname+ "' is not known.","",classerr,false,false);
    }
  return typep;
}






//--------------------------------------------------------------------------------
//---- CfpFactory::namehint() ----------------------------------------------------
//--------------------------------------------------------------------------------

string CfpFactory::namehint(string inname)
{
  // display a hint about name
  // this function is called when the user refers to an unknown class or variable in some scope.
  //  BUT even though the name is not known in scope, the user may be refering to soemthing out of scope,
  //  so we want to give them a hint about this.
  // do this by searching all classes, and class members
  CfpClass *classp=NULL;
  map<string, CfpClass *>::iterator p;
  string hintstring="";

  // we need to make sure the latest spawners have registered
  updatespawners();

  // first check if it is a class
  p=classmanager.begin();
  while (p!=classmanager.end())
    {
      classp=p->second;
      if (JRsimiliarstr(inname,classp->getname()))
	hintstring+="         class '"+classp->getname()+"' from package '" +classp->getpackagename()+"'.\n";
      p++;
    }

  // now check if it matches any class members
  p=classmanager.begin();
  while (p!=classmanager.end())
    {
      classp=p->second;
      hintstring+=classp->namehint(inname);
      p++;
    }

  return hintstring;
}






//--------------------------------------------------------------------------------
//---- Cfpfactory::preinterpret() ------------------------------------------------
//--------------------------------------------------------------------------------

void CfpFactory::preinterpret()
{
  // tell all registered classes that we are about to begin interpretation
  map<string, CfpClass *>::iterator p=classmanager.begin();
  CfpClass *classp;

  // we need to make sure the latest spawners have registered
  updatespawners();

  while (p!=classmanager.end())
    {
      classp=p->second;
      classp->preinterpret();
      p++;
    }
}



//--------------------------------------------------------------------------------
//---- Cfpfactory::postinterpret() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpFactory::postinterpret()
{
  // tell all registered classes that we are about done interpretation
  map<string, CfpClass *>::iterator p=classmanager.begin();
  CfpClass *classp;

  // we need to make sure the latest spawners have registered
  updatespawners();

  while (p!=classmanager.end())
    {
      classp=p->second;
      classp->postinterpret();
      p++;
    }
}




//--------------------------------------------------------------------------------
//---- Cfpfactory::postinterpret() -----------------------------------------------
//--------------------------------------------------------------------------------

void CfpFactory::addalias(string aliastr,string keyword,bool casesensitive=true)
{
  // eventually we want to migrate this functionality to within the factory, but
  // for now the configpp class handles it.

  // we need to make sure the latest spawners have registered
  updatespawners();

  if (configppp!=NULL)
    configppp->addalias(aliastr,keyword,casesensitive);
}







// }
// Namespace close-brace
