//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  Link.C
//  This file contains the class methods for the Link class.
//   see header file (.h) for description.
//
//


#include "misc.def"
#include "link.h"
#include "physobject.h"
#include "chain.h"

#include <math.h>

#define LK_FIXED      0
#define LK_REVOLUTE   1






Link:: Link(const char *linkname,World *worldp)
  : Physicalobject(linkname, worldp, NO)
{
  // This is a simple constructor which just initializes a default Link

  // Initialize there is no veloc, no accel., no torque applied, etc.
  setdefaults();
}



void Link:: setdefaults()
{
  // default type
  joint_type=LK_REVOLUTE;
  
  // Set frictions
  viscous=0.0;
  coulomb=0.0;

  // default dh
  a_length=0.0;
  alpha_twist=0.0;
  d_offset=0.0;
  theta_angle=0.0;

  // default limits
  poslimit=0;
  neglimit=0;
  K_stiffness=0;
  Y_damper=0;

  dtheta_angle=0.0;
  ddtheta_angle=0.0;
  last_theta_angle=0.0;

  extTorq=0.0;
  userTorq=0.0;
  alltorq=0.0;
  biastorq=0.0;

  springlimitko= DEFAULT_SPRINGLIMIT;
  hooklimit=0.0;
  overextended=NO;

  goal_angle=0.0;
  gotogoal=NO;
  pid_p=0.0;
  pid_i=0.0;
  pid_d=0.0;
  pid_limit=999999; // no limit on torq
  lastdist=0;

  // center of mass of link from center of mass of base physical object
  Com.fill(0.0);
  hinge_d=Com;
  hinge_c.fill(0.0); // should eventually be (length,0,0)
}


void Link::setdh(double length, double twist, double offset, double theta)
{
  a_length=length;
  alpha_twist=twist;
  d_offset=offset;
  theta_angle=theta;
}




void Link::setphys(double inlength,double inmass,double incomlength)
{
  // set the center of mass
  // center of mass of link from center of mass of base physical object

  // defaults to this inertia
  inertia(0,0)=inmass*inmass;
  inertia(1,1)=inmass;
  inertia(2,2)=inmass;

  Com(0)=incomlength;
  hinge_d=Com;
  hinge_c(0)=inlength;
  kilograms=inmass;
}


void Link:: done_construct()
{
  //  call Computevals() to initialize Freeman dynamics vars
  hinge_d=Com;

  Allocatevals();
  Computevals();
}









































































Link:: ~Link()
{
  // The destructor for the Link class
}


percision Link:: tauforce()
{
  // Return the torque applied at the joint due to external CNS control
  percision torq;
  torq=0.0;
  
  // this is torq from contacts and torq from user control
  torq=extTorq+userTorq;
  return torq;
}


int Link:: isfixed()
{
  // This is a query about whether this is a fixed, unrotatable link,
  //  which is indicated by a designated poslimit value
  if (joint_type==LK_FIXED)
    return(YES);
  else
    return(NO);
}


percision Link:: taufriction()
{
  // Return the torque applied at the joint due to friction
  percision torq;
  
  // 3/12/97 not sure about friction stuff
  if (dtheta_angle>0)
    torq=(-1.0*coulomb) - (viscous*dtheta_angle);
  else if (dtheta_angle<0)
    torq=coulomb - (viscous*dtheta_angle);
  else
    torq=0.0;
  
  return torq;
}


void Link:: setangle(percision ang)
{
  // Set the new angle of the link,
  //  and save a copy of the last value, which is needed during
  //  leg overextension past its limit.
  
  // handle angles over the 360 degrees
  while (ang>M_PI*2)
    ang-=(M_PI*2);
  while (ang<M_PI*(-2.0))
    ang+=(M_PI*2);
  
  last_theta_angle=theta_angle;
  theta_angle=ang;
}


void Link:: setextangle(percision ang)
{
  // Force a new angle and reset veloc and accel to 0.0
  //   this is for testing purposes
  //  and save a copy of the last value, which is needed during
  //  leg overextension past its limit.
  overextended=NO;
  goal_angle=ang;
}


void Link:: addtorque(float trq)
{
  // Add some contact applied torque to the link
  extTorq+=trq;
}


void Link:: setusertorque(percision trq)
{
  // set the user torque
  userTorq=trq;
}

void Link:: setexttorque(percision trq)
{
  // set the user torque
  extTorq=trq;
}


percision Link:: tauoverextend()
{
  // Return the torque applied at the joint due to any overextension
  percision overext=0.0;
  percision torq=0.0;
  percision deltatheta;
  
  // If this is a fixed link, it can't overextend
  if (isfixed()==YES)
    return(0.0);
  
  if (theta_angle<=poslimit && theta_angle>=neglimit)
    {
      // Not past positive or negative limit
      overextended=NO;
      return 0.0;
    }
  
  if (overextended==NO)
    {
      // They have just broken the limit, so we need to set the hook pos.
      // First a test to see if it is a manual readjustment,
      //  i.e. if last_theta_angle is also out of bounds
      if (last_theta_angle>poslimit || last_theta_angle<neglimit)
	{
	  // manual readjustment, so set hook at goal position
	  // this is more of a hack, and should be replaced with a
	  //  dedicated controller.
	  hooklimit=poslimit;
	}
      else
	hooklimit=last_theta_angle;
      overextended=YES;
    }
  
  // apply spring-damper equation from Freeman Thesis (89) p.36
  deltatheta=theta_angle-hooklimit;
  torq=(-1.0*K_stiffness*deltatheta)-(Y_damper*dtheta_angle);
  return torq;
}





void Link::movie_load(FILE *moviefp)
{
  // load all data for a link
  // normally we would call Physicalobject::movie_load()
  //  but we don't need everything saved, so instead we do it here manually

  (shape->grobp)->readlocation(moviefp);
  (shape->grobp)->readorientation(moviefp);
  Jvecload(shape->endlocation,moviefp,3);

  Jvecload(extF,moviefp,3);
  //  Jvecload(extN,moviefp,3);
  //  Jvecload(velocity,moviefp,3);
  //  Jvecload(userextF,moviefp,3);
  //  Jvecload(userextN,moviefp,3);
  
  // now angles, etc.
  fread(&theta_angle,sizeof(theta_angle),1,moviefp);
  fread(&dtheta_angle,sizeof(dtheta_angle),1,moviefp);
  fread(&ddtheta_angle,sizeof(ddtheta_angle),1,moviefp);
  
  // new to version 3.0
  fread(&extTorq,sizeof(extTorq),1,moviefp);
  //  fread(&userTorq,sizeof(userTorq),1,moviefp);
  //  fread(&biastorq,sizeof(biastorq),1,moviefp);
  //  fread(&goal_angle,sizeof(goal_angle),1,moviefp);
}



void Link::movie_save(FILE *moviefp)
{
  // save all data for a link
  // normally we would call Physicalobject::movie_save()
  //  but we don't need everything saved, so instead we do it here manually


  (shape->grobp)->writelocation(moviefp);
  (shape->grobp)->writeorientation(moviefp);
  Jvecsave(shape->endlocation,moviefp,3);

  Jvecsave(extF,moviefp,3);
  //  Jvecsave(extN,moviefp,3);
  //  Jvecsave(velocity,moviefp,3);
  //  Jvecsave(userextF,moviefp,3);
  //  Jvecsave(userextN,moviefp,3);

  fwrite(&theta_angle,sizeof(theta_angle),1,moviefp);
  fwrite(&dtheta_angle,sizeof(dtheta_angle),1,moviefp);
  fwrite(&ddtheta_angle,sizeof(ddtheta_angle),1,moviefp);

  // new to version 3.0
  fwrite(&extTorq,sizeof(extTorq),1,moviefp);
  //  fwrite(&userTorq,sizeof(userTorq),1,moviefp);
  //  fwrite(&biastorq,sizeof(biastorq),1,moviefp);
  //  fwrite(&goal_angle,sizeof(goal_angle),1,moviefp);
}







float Link:: taubuiltin()
{
  // Return the torque applied at the joint due to builtin controller
  float torq;
  float dist,distd;
  float pcomponent,icomponent,dcomponent;
  float totalappliedtorq;
  
  dist=goal_angle-theta_angle;
  distd=(dist-lastdist)*100;
  lastdist=goal_angle-theta_angle;

  pcomponent=dist*pid_p;
  icomponent=0.0;
  dcomponent=distd*pid_d;

  totalappliedtorq=pcomponent+icomponent+dcomponent;

  // now check torq limit and constrain it
  if (totalappliedtorq>pid_limit)
    totalappliedtorq=pid_limit;
  else if (-1*totalappliedtorq>pid_limit)
    totalappliedtorq=pid_limit*-1;
   
  torq=totalappliedtorq;
  return torq;
}








// NOTE;
// Link::Computevals() is now contained in dynamics.C;









