//  Jesse Reichler, January 1996
//  Fred Delcomyn
//
//  link.h
//  This file contains the class definition for the Link class,
//   a component class of the Chain class.
//  Open chains are made up of links described in Denavit/Hartenberg format
//  The Chain class is defined in a separate file.
//  
//

#ifndef Link_defined
#define Link_defined

#include "misc.def"
#include "matrixinc.h"
#include "physobject.h"




class Link : public Physicalobject
{
  friend class Rot;
  friend class Robot;
  friend class Chain;
public:
  Link(const char *linkname,class World *worldp);
  ~Link();
public:
  void setdefaults();
  void setdh(double length, double twist,double offset, double theta);
  void setphys(double inlength,double inmass,double incomlength);
  void done_construct();
private:
  void Allocatevals();	// allocate space for inertia matrix, etc.
  void Computevals();	// construct the inertia matrix, etc.;
private:
  // Denavit-Hartenberg parameters, note that theta_angle;
  //  is a dynamic kinematic description, not used in describing; the structure of the link/joint.
  percision a_length;
  percision alpha_twist;
  percision d_offset;
  percision theta_angle;
  percision last_theta_angle;
  
  // Veloc and Accel of the joint(link) variable theta_angle;
  percision dtheta_angle;
  percision ddtheta_angle;
  
  // Limits of the joint, in radians;
  // from legsim2;
  percision poslimit;
  percision neglimit;
  percision springlimitko;   // Nm/rad (see legsim2);
  percision hooklimit;
  int overextended;

  // A goal angle for internal, simplistic controller;
  int gotogoal;
  percision goal_angle;
  float pid_p,pid_i,pid_d;
  float pid_limit;
  float lastdist;
  
  // New overextension spring stuff;
  percision K_stiffness;
  percision Y_damper;
  
  // Frictions at the joint;
  percision coulomb;
  percision viscous;
		
  // Joint type, Fixed or Revolute;
  int joint_type;

  // Inertia Tensor for the link, (the moment of inertia);
  //  defined as per [Craig, 1986, p.190], and using E.Weltons matrix stuff;
  perVectorR3 Seg;
  percision force;
  percision extTorq;
  percision userTorq;
  percision alltorq;
  percision biastorq;

private:
  // The *new* hinge distance dynamics variable, set in legcone shape
  perVectorR3 hinge_d;
  perVectorR3 hinge_c;

private:
  // The new dynamics data;
  perMatrix DI;
  // coordinate transforms;
  perMatrix DX;
  perMatrix DA;
  // Forward Kinemats;
  perVector Dw;
  perVector Dv;
  perVector Do;
  perVector DB;
  perVector Df;
  // Backward dynamits;
  perMatrix DISTAR;
  perVector DF;
  perVector DBSTAR;
  perMatrix DN;
  perVector DHT;
  percision Dtau;
  //Forward dynamits;
  perVector Da;
  perMatrix inertiaIH; // inertia w.r.t. link hinge point (with prev. link)
public:
  float getlength() {return ((Legcone *)shape)->getlength();};
  float getangle() {return theta_angle;};
  float getgoalangle() {return goal_angle;};
  void setangle(percision ang);
  void setgoalangle(float goalang) {goal_angle=goalang; lastdist=goal_angle-theta_angle; };
  void setpid_p(float p) {pid_p=p;};
  void setpid_i(float i) {pid_i=i;};
  void setpid_d(float d) {pid_d=d;};
  void setpid_limit(float t) {pid_limit=t;};
  float getpid_p() {return pid_p;};
  float getpid_i() {return pid_i;};
  float getpid_d() {return pid_d;};
  float getpid_limit() {return pid_limit;};
  float geterror() {return theta_angle-goal_angle;};
  float gettorq() {return (float)Dtau;};
public:
  void setextangle(percision ang);
public:
  percision tauforce();
  percision taufriction();
  float taubuiltin();
  percision tauoverextend();
  int isfixed();
  void addtorque(float trq);
  void setusertorque(percision trq);
  void setexttorque(percision trq);
public:
  percision getusertorque(){return userTorq;};
  percision getalltorque(){return userTorq;};
  percision getexttorque(){return extTorq;};
  percision getfelttorque(){return biastorq;};
public:
  virtual void movie_load(FILE *moviefp);
  virtual void movie_save(FILE *moviefp);
public:
  void sethinges(float len){hinge_d(0)=len/2;hinge_c(0)=len;};
  void sethinge_d(float hx,float hy,float hz){hinge_d(0)=hx;hinge_d(1)=hy;hinge_d(2)=hz;};
  void sethinge_c(float hx,float hy,float hz){hinge_c(0)=hx;hinge_c(1)=hy;hinge_c(2)=hz;};
};



#endif





