//  Jesse Reichler, February 1996
//  Fred Delcomyn
//
//  rot.C
//  This file contains the class methods for the Rot class,
//   see heade file (.h) for description
//  
// The Rot class is used to transform the Dennavit Hartenbberg parameters
//  between link frames.  This means that the use of Rot will be in calls
//  of the form ROT[i][i-1], ROT[i][i+1], ROT[i][0], or ROT[0][i].
// The ROT values are related to the links in a single chain.
//  each ROT class is attached to a specific chain, in addition to the
//  function calls returning the ROT values for a specific transform,
//  we must support an update() function to recalculate the matrices once the
//  link positions have changed.
// NOTE that the more sensible 4x4 DH transform subsumes both ROT and ROTp
//  but for now we want to keep consistency with legsim so we are breaking
//  the two up.
// NOTE: frame 0 is base, frame N is ground
//


#include "matrixinc.h"
#include "rot.h"
#include "link.h"
#include "base.h"
#include "chain.h"
#include "robot.h"
#include "jessdh.h"

#include <stdio.h>




void Rot::allocate(Chain *ichainp, World *iworldp)
{
  // Allocate the internal space we need for the ROT data.
  // NOTE: for computational efficiency, we sometimes refer to indices which
  //  are outside the bounds of the number of links.
  //  We must be careful to allocate for this case.
  
  // Many ways to implement this class from space efficient to time efficient
  //  here i don't worry so much about space requirements N*N in # of links
  //  this can be reduced to 4*N and will be if the move to PHASE 2 doesn't
  //  completely replace this.
  // ROTdata is defined as:   perMatrix ***ROTdate;
  int count,count1,count2;
  int reslinks;
  
  // Store the internal pointer to our chain&world who we are attached to
  chainp=ichainp;
  worldp=iworldp;
  
  reslinks=chainp->links+3;
  
  // Allocate space for reslinks links
  // First ROTdataX
  ROTdataX=new perMatrix* [reslinks];
  for (count=0;count<reslinks;++count)
    ROTdataX[count]=new perMatrix [reslinks];
  
  // Now ROTdataT
  ROTdataT=new perMatrix* [reslinks];
  for (count=0;count<reslinks;++count)
    ROTdataT[count]=new perMatrix [reslinks];
  
  // Now initialize them all - this really isn't nesc.
  //  since Welton's class allows us to delete uninitialized matrices
  //  without memory corruption, but we do it nontheless for completeness.
  
  // First ROTdataX
  for (count1=0;count1<reslinks;++count1)
    {
      for (count2=0;count2<reslinks;++count2)
	ROTdataX[count1][count2]=perMatrix(6,6);
    }
  
  // First ROTdataT
  for (count1=0;count1<reslinks;++count1)
    {
      for (count2=0;count2<reslinks;++count2)
	ROTdataT[count1][count2]=perMatrix(4,4);
    }
  
  initialize();
}






void Rot::deallocate()
{
  // Reverse the allocation process and free up the memory
  int count;
  int reslinks;
  
  reslinks=chainp->links+3;
  
  // Note we must delete inner matrices first
  //  the delete[] expression says to delete the array
  
  for (count=0;count<reslinks;++count)
    delete [] ROTdataX[count];
  delete [] ROTdataX;
  
  for (count=0;count<reslinks;++count)
    delete [] ROTdataT[count];
  delete [] ROTdataT;
}









void Rot::initialize()
{
  perMatrix E6(6,6);
  perMatrix E4(4,4);
  perMatrix E3(3,3);
  perMatrix mat3a(3,3);
  perMatrix mat3b(3,3);
  perMatrix mat3c(3,3);
  perMatrix mat3(3,3);
  perVector tempv3c(3,ColV);
  int lastlink,footlink,firstlink;
  Link *linkp;
  int count,count1,count2;
  int reslinks;
  
  reslinks=chainp->links+3;
  
  //  All i to i should be identity matrix
  E6.fill(0.0);
  E6(0,0)=E6(1,1)=E6(2,2)=E6(3,3)=E6(4,4)=E6(5,5)=1.0;
  E4.fill(0.0);
  E4(0,0)=E4(1,1)=E4(2,2)=E4(3,3)=1.0;
  E3.fill(0.0);
  E3(0,0)=E3(1,1)=E3(2,2)=1.0;
  
  // First ROTdataX initialize
  for (count=0;count<reslinks;++count)
    ROTdataX[count][count]=E6;
  
  // Now ROTdataT initialize
  for (count=0;count<reslinks;++count)
    ROTdataT[count][count]=E4;
  
  // Now set any that remain constant
  lastlink=(chainp->links-1)+2; // we add 2 to all link addresses
  footlink=lastlink+1;
  
  // Foot to Last Link stays constant
  
  // NOTE: 12/9/96 this is a guess about what value should be used here,
  //  I think more appropriate is [lastlinklength 0 0]
  mat3a=E3;
  linkp=(chainp->link)[lastlink-2];
  tempv3c=(linkp->hinge_c);
  ROTdataX[footlink][lastlink]=JessmakeX(mat3a,tempv3c);
  ROTdataX[lastlink][footlink]=JessmakeXi(mat3a,tempv3c);
  ROTdataT[footlink][lastlink]=JessmakeT(mat3a,tempv3c);
  ROTdataT[lastlink][footlink]=~ROTdataT[footlink][lastlink];
  
  // Body to hip stays constant
  mat3.fill(0.0);
  
  // incorporate hip starting orientation
  mat3=JessmakeA(0.0,0.0,chainp->hip_alpha_twist,chainp->hip_theta_angle);
  tempv3c=(chainp->hip_distance);
  
  ROTdataX[RII_HIP][RII_BODY]=JessmakeX(mat3,tempv3c);
  ROTdataT[RII_HIP][RII_BODY]=JessmakeT(mat3,tempv3c);
  ROTdataX[RII_BODY][RII_HIP]=JessmakeXi(mat3,tempv3c);
  ROTdataT[RII_BODY][RII_HIP]=~ROTdataT[RII_HIP][RII_BODY];
}




// ------------------------------------------------------------------------------------
// Rot::update()
// ------------------------------------------------------------------------------------


void Rot::update()
{
  // In this case, the ROTdata is already allocated, we just need to update
  //  the transformation matrics based on the new values of the link
  //  variables. Rot matrics are X matrices (whose upper left 3x3 is A)
  int count;
  int i,j;
  perMatrix mat6(6,6);
  perMatrix Amat(3,3);
  perMatrix mat3(3,3);
  perMatrix mat3b(3,3);
  perMatrix mat3c(3,3);
  perMatrix zero3(3,3);
  perMatrix tempv3c(3,ColV);
  perMatrix tempv3c2(3,ColV);
  Link *linkp;
  int firstlink,lastlink,footlink;
  Base *basep;
  
  // Set indices for last and foot links
  firstlink=0+3;   // we skip over 3 (earth,body,hip)
  lastlink=(chainp->links-1)+2;
  footlink=lastlink+1;
  // init
  basep=(chainp->robotp)->base;
  zero3.fill(0.0);
  
  // First compute the Body-Earth Transform
  // which is based on the Base DeAb A transform matrix
  
  // orig 5/9/97
  mat3=(basep->DbAe);
  
  tempv3c=(basep->locationval());
  ROTdataX[RII_BODY][RII_EARTH]=JessmakeX(mat3,tempv3c);
  ROTdataT[RII_BODY][RII_EARTH]=JessmakeT(mat3,tempv3c);
  ROTdataX[RII_EARTH][RII_BODY]=JessmakeXi(mat3,tempv3c);
  ROTdataT[RII_EARTH][RII_BODY]=~ROTdataT[RII_BODY][RII_EARTH];

  for (i=firstlink;i<=lastlink;++i)
    {
      // to i from i-1
      linkp=(chainp->link)[i-2];
      Amat=JessmakeA(linkp->a_length,linkp->d_offset,linkp->alpha_twist,linkp->theta_angle);
      tempv3c=Jessmakeci(linkp->a_length,linkp->d_offset,linkp->alpha_twist,linkp->theta_angle);
      
      ROTdataX[i][i-1]=JessmakeX(Amat,tempv3c);
      ROTdataT[i][i-1]=JessmakeT(Amat,tempv3c);
      ROTdataX[i-1][i]=JessmakeXi(Amat,tempv3c);
      ROTdataT[i-1][i]=~ROTdataT[i][i-1];
    }

  // Compute transforms to and from Earth and all links
  // First do hips
  ROTdataX[RII_EARTH][RII_HIP]=ROTdataX[RII_EARTH][RII_BODY]*ROTdataX[RII_BODY][RII_HIP];
  ROTdataT[RII_EARTH][RII_HIP]=ROTdataT[RII_EARTH][RII_BODY]*ROTdataT[RII_BODY][RII_HIP];
  ROTdataX[RII_HIP][RII_EARTH]=ROTdataX[RII_HIP][RII_BODY]*ROTdataX[RII_BODY][RII_EARTH];
  ROTdataT[RII_HIP][RII_EARTH]=ROTdataT[RII_HIP][RII_BODY]*ROTdataT[RII_BODY][RII_EARTH];
  
  for (i=firstlink;i<=footlink;++i)
    {
      // to EARTH from i, calculated recursivly using the prev.
      ROTdataX[i][RII_EARTH]= ROTdataX[i][i-1] * ROTdataX[i-1][RII_EARTH];
      ROTdataT[i][RII_EARTH]= ROTdataT[i][i-1] * ROTdataT[i-1][RII_EARTH];
      ROTdataX[RII_EARTH][i]= ROTdataX[RII_EARTH][i-1]*ROTdataX[i-1][i];
      ROTdataT[RII_EARTH][i]= ROTdataT[RII_EARTH][i-1]*ROTdataT[i-1][i];
    }

  // We also need to add [BODY][1] and [1][BODY], some transforms which 
  //  bypass the hip
  ROTdataX[RII_BODY][firstlink]= ROTdataX[RII_BODY][RII_HIP] * ROTdataX[RII_HIP][firstlink];
  ROTdataT[RII_BODY][firstlink]= ROTdataT[RII_BODY][RII_HIP] * ROTdataT[RII_HIP][firstlink];
  ROTdataX[firstlink][RII_BODY]= ROTdataX[firstlink][RII_HIP] * ROTdataX[RII_HIP][RII_BODY];
  ROTdataT[firstlink][RII_BODY]= ROTdataT[firstlink][RII_HIP] * ROTdataT[RII_HIP][RII_BODY];

  //  verify();
}



















// ------------------------------------------------------------------------------------
// New dynamics transforms
// ------------------------------------------------------------------------------------


perMatrix Rot::DXA33(int indexto,int indexfrom)
{
  // Return the 3x3 transform matrix that will transform coordinates;
  //  to frame indexto from frame indexfrom.;
  perMatrix transmat(3,3);
  
  // correct index from -2 to N to 0 N+2
  indexto+=2;
  indexfrom+=2;

  // A[to][from] is just the upperleft 3x3 of X[to][from]
  transmat=Jessup33(ROTdataX[indexto][indexfrom]);
  return transmat;
}



perMatrix Rot::DX66(int indexto,int indexfrom)
{
  // Return the 6x6 transform matrix that will transform coordinates
  //  to frame indexto from frame indexfrom.
  perMatrix transmat(6,6);
  
  // correct index from -2 to N to 0 N+2
  indexto+=2;
  indexfrom+=2;
  
  transmat=ROTdataX[indexto][indexfrom];
  return transmat;
}


perMatrix Rot::DX66i(int indexto,int indexfrom)
{
  // Return either ~DX66(to,from),  or DX(from,to), depending on how we read freeman's thesis

  perMatrix transmat(6,6);

  // alternate way
  transmat=DX66(indexfrom,indexto);
  
  return transmat;
}


perMatrix Rot::DT44(int indexto,int indexfrom)
{
  // Return the 4x4 transform matrix that will transform coordinates to frame indexto from frame indexfrom.
  perMatrix transmat(4,4);
  
  // correct index from -2 to N to 0 N+2
  indexto+=2;
  indexfrom+=2;
  
  transmat=ROTdataT[indexto][indexfrom];
  return transmat;
}


perMatrix Rot::DTA33(int indexto,int indexfrom)
{
  // Return the 3x3 orientation matrix that will transform coordinates
  //  to frame indexto from frame indexfrom.
  perMatrix transmat(3,3);
  int i,j;
  
  // correct index from -2 to N to 0 N+2
  indexto+=2;
  indexfrom+=2;
  
  for (i=0;i<=2;i++)
    {
      for (j=0;j<=2;j++)
	transmat(i,j)=ROTdataT[indexto][indexfrom](i,j);
    }
  

  return transmat;
}



perVector Rot::DTp(int indexto,int indexfrom)
{
  // Translate the point between frames
  //  to frame indexto from frame indexfrom.
  perVector tempv1(4,ColV);
  perVector retvec;
  
  // correct index from -2 to N to 0 N+2
  indexto+=2;
  indexfrom+=2;
  
  // contstruct the 0,0,0 location vector (Spong 1989, p. 48)
  tempv1.fill(0.0);
  tempv1(3)=1.0;
  
  retvec=ROTdataT[indexto][indexfrom]*tempv1;
  return retvec;
}



perMatrix Rot::DXcongru(int indexto,int indexfrom,perMatrix N)
{
  // Perform a congruency transform on an inertia matrix
  perMatrix retmat(6,6);

  retmat=DX66i(indexfrom,indexto)*N*DX66(indexfrom,indexto);
  return retmat;
}






void Rot::verify()
{
  // verify that matrices and inverses are okay
  int i;
  int firstlink,lastlink,footlink;
  perMatrix m1(6,6);
  perMatrix m2(6,6);
  perMatrix m3(6,6);

  printf("$Performing verificaton:\n");
  
  firstlink=0+3;   // skip earth,body,hip
  lastlink=(chainp->links-1)+2;
  footlink=lastlink+1;
  for (i=1;i<=footlink;++i)
    {
      m1=ROTdataX[i-1][i];
      m2=ROTdataX[i][i-1];
      m3=m1*m2;

      printf("$should be indentity (%d,%d):\n",i,i-1);
      m3.print();
      getchar();
    }

  printf("with earth:\n");
  for (i=1;i<=lastlink;++i)
    {
      m1=ROTdataX[RII_EARTH][i];
      m2=ROTdataX[i][RII_EARTH];
      m3=m1*m2;

      printf("$should be indentity (%d,%d):\n",RII_EARTH,i-1);
      m3.print();
      getchar();
    }
}










