// configpp.h
// This file describes the main class which handles all Configpp parsing.
// It is basically the core Configpp interpreter.

#ifndef INC_configpph
#define INC_configpph

#include "cfpinc.h"

//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace Configpp{







//--------------------------------------------------------------------------------
//---- Configpp -------------------------------------------------------------------
//--------------------------------------------------------------------------------

//: The Configpp class is the main Configpp parser/interpreter.
// It requires the helper class CfpFactory, which manages the
// CfpObject class library (which defines all object classes), including user extensions.

class Configpp
{
  friend class CfpWrap;						// the class wrapper
private:
  CfpWrap *cfpwrapp;
  CfpFactory *factoryp;						// the class factory pointer
  CfpLexer *lexerp;						// the lexer pointer for extracting tokens
  CfpTimerManager *timermanagerp;
  list<string> searchdirs;					// list of directories to search when opening files
  JRAliasmap aliases;						// some aliases for class/variable/keywords;
  deque<CfpObject *> scopestack;				// our current scope
  map<string,CfpObject *> objtable;				// map of existing objects by full trail name
  int largc;							// local copy of argc
  char **largv;							// local copy of argv
  bool diddone;							// to keep track of whether user executed done()
  int idcounter;						// counter for generating unique idnums for objs
public:
  JRrand jrandhelper;
private:
  string appname;
  bool Debugmode;						// display debug messages?
  bool Warningmode;						// display warnings?
  bool Hintmode;						// display hints?
  int errors;							// number of errors so far
  int Maxerrors;						// max number of syntax errors before exiting
  CfpLoc Hintloc;						// hint about where we think the last error was
  CfpVal GlobalReturnValue;					// return value from last func call or obj inst.
  CfpParserModeType Parsermode;					// syntax check or actually run code
private:
  bool Allowpounds;						// do we allow pound directives?
public:
  Configpp(string inappname,CfpFactory *infactoryp,int argc=0,char **argv=NULL);	// constructor
  ~Configpp();
public:
  void welcome();						// print a welcome message
  void commandline(int argc,char **argv);			// process configpp commandline flags
  void checkforflags(char *word);				// process flags in word
  void static show_about();					// print some info about Configpp
  void static show_commandline_options();			// print commandline options
public:
  bool pushfile(string filename);				// push a file on inputsource stack
  bool pushstring(string parsestring,string name="");		// push a string on inputsource stack
  bool parsefile(string filename);				// push a file on inputsource stack, parse
  bool parsestring(string parsestring,string name="");		// push a string on inputsource stack, parse
  void done();							// when we are all done parsing
  bool run();							// run the GUI and the simulation
public:
  void addsearchdir(string dir);				// add a directory to the search-dir list
public:
  string getappname() {return appname;};
  CfpToken *getcurrentokenp() {return lexerp->getcurtokenp();};				// accessor
  bool debugmode() {return Debugmode;};							// accessor
  bool hintmode() {return Hintmode;};							// accessor
  CfpLexer *getlexerp() {return lexerp;};						// accessor
  CfpLoc *gethintlocp() {return &Hintloc;};						// accessor
  void sethintloc(CfpLoc *inlocp){Hintloc=*inlocp;};					// accessor
  void sethintloc(CfpToken *tokep){Hintloc=*(tokep->getlocp());};			// accessor
  CfpFactory *getfactoryp(){return factoryp;};						// accessor
  CfpTimerManager *gettimermanagerp(){return timermanagerp;};				// accessor
  string getcurfilename(){return lexerp->getcurfilename();};				// accessor
  long getcurpos(){return lexerp->getcurpos();};					// accessor
  CfpInputSource *getcursource(){return lexerp->getcursource();};			// accessor
public:
  void debug(string debugstring);							// print debugstring
  void warning(string warnstring, CfpLoc *wloc=NULL);					// print warning
  void announce(string warnstring);							// print announcement
  void error(string errstring,string hintstr,CfpErrorCode etype, bool fatalflag=false);	// display an error
  void handlecomment(string commentstr,string sourcename);				// throwaway a comment
  string semihint();					// return a string hint if we guess CR intended semicolon
  string namehint(string name);								// return a hint about name;
  bool getwarningmode() {return Warningmode;};
  bool getdebugmode() {return Debugmode;};
public:
  fstream *inFilestream(string filename, const string ext="");	// find & open the file stream for input
  fstream *inFilestream(string filename, string &outfilename, const string ext="");
  fstream *outFilestream(string filename, string &outfilename, const string ext="");
  fstream *outappFilestream(string filename, string &outfilename, const string ext="");
public:
  bool parse();							// do the parsing, return false if we hit errors
public:
  string alias(string wordstr);					// return what wordstr is an alias of
  void addalias(string aliasstr,string keyword,bool casesensitive=true);	// add a new alias to the list of aliases
private:
  void loadrc() {loadrcenv();loadrcfiles();};		// load file and environment configs
  void loadrcenv();					// load environment config
  void loadrcfiles();					// load file configs
  void foundations(int *argcp,char **argv);		// create and register base factory classes and objs needed
  void delete_objects();				// delete objects we have created
  void destructobjtree(CfpObject *objp);		// delete objects we have created
  void tableremove(CfpObject *objp);			// remove obj from the objtable
private:
  void Loadlibs(int *argcp,char **argv);		// load all libs we know about
  void Loadlib(string libname,int &argc,char **argv) {;};	// load a given library of class info
public:
  void addtimer(CfpTimer *timerp){timermanagerp->addtimer(timerp);};			// add a new timer
private:
  void gosyntaxcheckmode();					// switch to syntax checking mode, tell user
  void processerror(CfpError *serr);				// display the error, and process it
  string scopestring();						// return a string version of the current scope
private:
  CfpTokenType parseconsume(bool retcr=false);			// throw away tokens until end of error
  bool Parsestatement();					// parse a single statement/block
  CfpTokenType internalParsestatement(bool retclosebrace);	// parse a single statement/block
  void Parseblock();						// parse a block of statements
  void Parsepound(CfpToken stoke);				// parse a pound directive
  void Parsealpha(CfpToken stoke);				// parse a statement that begins with word
  void Parseassignment(CfpObject *objp) {;};			// parse an assignment
  void Parseassignment(CfpVAccessor *varp,string vectoradstring="");			// parse an assignment
  string Parsevectoradstring();			                // read in a vector address string (ie "[1 2 3]")
  void Parsedeclaration(string classname,CfpClass *classp);	// parse a delcaration
  CfpVal Parsefunc(string funcname);				// parse a function call
  CfpVal Parsevalexp();						// parse a value expression
  CfpVal Parsevalue();						// parse a single value
  void Parsesemicolon();					// error if next token !=semicolon
public:
  bool isdefined(string symname);				// return true if symbol/obj/class is defined
  string classname(string objname,bool complain=true);		// return the class name of an object
  void runconstructor(CfpObject *objp);				// invoke the approp. lazy instantiator
  CfpVAccessor *lookupVarArg(string identifier,bool complain=true,bool skipcurrent=false);	// smart lookup the var or arg
  CfpVAccessor *lookupVariable(string varname,bool complain=true);	// lookup the variable and return pointer
  CfpFuncAccessor *lookupFunction(string funcname,bool complain=true);	// lookup the function and return pointer
  CfpObject *lookupObjtrail(string objtrail,bool complain=true);	// lookup the objtrail and return pointer
  CfpVal lookupVal(string identifier,bool complain=true);	// lookup the var/arg in scope and return value
  CfpVAccessor *lookupVAcc(string identifier,bool complain=true);	// smart lookup the (obj) var or arg;
  bool compatibletrail(string fulltrail,string candidatetrail);
private:
  CfpObject *createObject(string classname, string objname="");	// create a new object instantiation
  void completeObject(CfpObject *objp);				// complete the instantiation of the object
  void procAssignment(string varname, string varval);		// perform the assignment
public:
  bool allow_shorthandassignments() {return true;};             // turn this off for more strict parsing (see cfpparser.C);
  CfpVAccessor *findvaccessor(string name,int &index);
  string tokenreplacement(string instring);
  bool expressionval(string identifier,CfpVal &val);
public:
  CfpObject *adoptobj(CfpObject *cobjp);				// the object wants to be adopted
  CfpObject *inscopefinda(string typestr);				// find the nearest inscope obj which isa
  CfpObject *finda(string typestr);					// find any instantiated object which isa
  CfpObject *findchild(CfpObject *pobjp,string typestr,bool complain);	// find any instantiated object which isa
//  CfpObject *findchild(CfpObject *pobjp,string typestr,bool complain=true);	// find any instantiated object which isa
  CfpObject *enclosingparent(CfpObject *childobjp);			// find enclosing parent
  CfpObject *parentinstantiate(CfpObject *childobjp,string typestr,bool complain=true);	// instantiate parent
  void quit_request();									// quit program;
};




// }
// Namespace close-brace
#endif
