/***********************************************************************

 MatrixT.h
 MatrixT Class Member Function Definitions Implementation
 Copyright (C) 1993
        Lyle Hall 1993/94
        Eric C. Welton 1993/94
        Beckman Institute - Urbana, Illinois
        For Neuronal Pattern Analysis Group and Hexapod Project

        These libraries are free and may be redistributed and/or modified,
        provided the above notice and this paragraph are included.
        These libraries may not be sold and are distributed without
        any warranty or implication of fitness for a purpose.

 General Information

        The MCLASSNAME library provides fundamental matrix operations.

***********************************************************************/

#ifndef MATRIX_IMPLEMENTATION_HDR
#define MATRIX_IMPLEMENTATION_HDR

#include <MatrixT.h>
#include <math.h>




//======================================================================
//
// #define flag status at this point should be:
//
// DEBUG_SINK              correct from Matrix.h
// SUBSCRIPTION_OFFSET     correct from Matrix.h
// DEFAULT_PRECISION       correct from Matrix.h ('float' or 'double')
// DEF_PREC_VECTOR_NAME    correct from Matrix.h ('DVector' or 'FVector')
// DEF_PREC_VECTOR_PREFIX  correct from Matrix.h ('D' or 'F')
// __Concat2               correct from Matrix.h
// _Concat2                correct from Matrix.h
// __Concat3               correct from Matrix.h
// _Concat3                correct from Matrix.h
//
// __DONT_DEFINE_MATRIX_TEMPLATE should be defined by any user
//                               or subclassing files prior to
//                               the inclusion of this file.
//
// TEMPLATE                undef, will def here is DONT... flag undef
// ELT                     undef, will def here is DONT... flag undef
// VCLASSNAME              undef, will def here is DONT... flag undef
// MCLASSNAME              undef, will def here is DONT... flag undef
// VECT                    undef, will def here is DONT... flag undef
// MAT                     undef, will def here is DONT... flag undef
// TYPER                   undef, need to define here
//


// determine whether this is building a header or object source
#ifndef INLINE
#define INLINE inline
#endif


//======================================================================
//
// When the __DONT_DEFINE_MATRIX_TEMPLATE is undefined (e.g. the
// default situation) the following macros are defined.  These
// cause this file to implement the MCLASSNAME<TYPE> template
//
//       TEMPLATE    template<class EltType>
//       TYPE        EltType
//       VCLASSNAME  VectorT
//       MCLASSNAME  MatrixT
//       VECT        VCLASSNAME<TYPE>
//       MAT         MCLASSNAME<TYPE>
//
// If these are defined differently it is possible to
// implement a non-template version of the Vector data type.
// This is done, for example, by the DMatrix header file, which
// defines the above as follows (and sets the __DONT... flag)
//
//       TEMPLATE
//       TYPE        double
//       VCLASSNAME  DVector
//       MCLASSNAME  DMatrix
//       VECT        DVector
//       MAT         DMatrix
//

#ifndef __DONT_DEFINE_MATRIX_TEMPLATE
#pragma interface
#define TEMPLATE template<class EltType>
#define TYPE EltType
#define VCLASSNAME VectorT
#define MCLASSNAME MatrixT
#define MAT MCLASSNAME<TYPE>
#define VECT VCLASSNAME<TYPE>
#else
#ifndef TEMPLATE
#define TEMPLATE
#endif
#ifndef MAT
#define MAT MCLASSNAME
#endif
#ifndef VECT
#define VECT VCLASSNAME
#endif
#endif

// this is used where a reference could also be used - in the
// future the template might be expanded to facilitate non
// basic types, or types with expensive copy functions - in
// this case we want to use references to pass arguments.
#define TYPER TYPE



//======================================================================
//
// Header build control macros
//

//========================= Error handling

#define ERROR(val,ret) { fprintf(stderr,"%s<> library error, line %d, file %s\n:::: error=%s\n",__CLASSNAMESTRING,__LINE__,__FILE__,#val); abort(); return ret; }


//========================= Element Access


// if the user has requested optimized element access, then, wherever
// possible we access the data directly, instead of through the subscription
// operator - the subscription operator should be used when debugging (i.e.
// turn off the optimization) since that provides the user with
// additional break points, error checking, and status information
#ifdef OPTIMIZE_ELEMENT_ACCESS
// yes, optimize
#define ROW_INDEX                ((__kk%width)+SUBSCRIPTION_OFFSET)
#define COL_INDEX                ((__kk/width)+SUBSCRIPTION_OFFSET)
#define FOR_EACH_ELEMENT         for(register int __kk=0; __kk<width*height; __kk++)
#define FOR_EACH_ROW_ELEMENT(i)  register int __upper=(i+1)*width;   \
                                    for(register int __kk=(i*width); __kk<__upper; __kk++)
#define FOR_EACH_COL_ELEMENT(i)  register int __upper=i+(height*width); \
                                    for(register int __kk=i; __kk<__upper; __kk+=width)
#define REFELT(m)          (m.data[__kk])
#define REFELTK(a,b)       (m.data[(a-SUBSCRIPTION_OFFSET)*width+(b-SUBSCRIPTION_OFFSET)])
#define REFELTKN(a,b)      (m.data[(a*width)+b])
#define ELT                (data[__kk])
#define ELTK(a,b)          (data[(a-SUBSCRIPTION_OFFSET)*width+(b-SUBSCRIPTION_OFFSET)])
#define ELTKN(a,b)         (data[(a*width)+b])

#else
// no, don't optimize
#define ROW_INDEX __ii
#define COL_INDEX __jj
#define FOR_EACH_ELEMENT                                    \
 for(register int ROW_INDEX=SUBSCRIPTION_OFFSET;            \
      ROW_INDEX<height+SUBSCRIPTION_OFFSET; ROW_INDEX++)    \
   for(register int COL_INDEX=SUBSCRIPTION_OFFSET;          \
         COL_INDEX<width+SUBSCRIPTION_OFFSET; COL_INDEX++)
#define FOR_EACH_ROW_ELEMENT(i)  COL_INDEX=i;               \
   for(register int ROW_INDEX=SUBSCRIPTION_OFFSET;          \
         ROW_INDEX<height+SUBSCRIPTION_OFFSET; ROW_INDEX++)

#define FOR_EACH_COL_ELEMENT(i)  ROW_INDEX=i;               \
   for(register int COL_INDEX=SUBSCRIPTION_OFFSET;          \
         COL_INDEX<width+SUBSCRIPTION_OFFSET; COL_INDEX++)

#define REFELT(m)       (m(ROW_INDEX,COL_INDEX))
#define REFELTK(m,a,b)  (m(a,b))
#define REFELTKN(m,a,b) (m(a+SUBSCRIPTION_OFFSET,b+SUBSCRIPTION_OFFSET))
#define ELT             ((*this)(ROW_INDEX,COL_INDEX))
#define ELTK(a,b)       ((*this)(a,b))
#define ELTKN(a,b)      ((*this)(a+SUBSCRIPTION_OFFSET,b+SUBSCRIPTION_OFFSET))

#endif


//========================= Absolute Value (we might optimize this later)
#define ABS(a) (a<0?-a:a)
#define ABSELT (ELT<0?-ELT:ELT)

//========================= Free-Store Management
#define GET_NEW_ELTS(name,num) name=new TYPE[num]
#define FREE_ELT_SPACE(name)   delete name

//========================= Vector Access Protection
#ifdef PROTECT_VECTOR_ACCESS
#define CHECK_FOR_VALID_VECTOR(ret)    \
             if(numelts<0) \
               ERROR(VectorT_Invalid_Access,ret);
#define CHECK_V_FOR_VALID_VECTOR(v,ret)               \
            if((&v)==0)                               \
               { ERROR(VectorT_Null_Operand,ret); }   \
            else                                      \
               { if(v.numelts<0)                      \
                  ERROR(VectorT_Invalid_Access,ret); }

#else
#define CHECK_FOR_VALID_VECTOR(ret)
#define CHECK_V_FOR_VALID_VECTOR(v,ret)
#endif

//========================= Matrix Access Protection
#ifdef PROTECT_MATRIX_ACCESS
#define CHECK_FOR_VALID_MATRIX(ret)    \
             if((width<0)||(height<0)) \
               ERROR(MatrixT_Invalid_Access,ret);
#define CHECK_M_FOR_VALID_MATRIX(m,ret)               \
            if((&m)==0)                               \
               { ERROR(MatrixT_Null_Operand,ret); }   \
            else                                      \
               { if((m.width<0)||(m.height<0))        \
                  ERROR(MatrixT_Invalid_Access,ret); }

#else
#define CHECK_FOR_VALID_MATRIX(ret)
#define CHECK_M_FOR_VALID_MATRIX(m,ret)
#endif

//========================= Ctor and Dtor Debugging Auxiliaries
#define ___CREATE_STRING(a) #a
#define __CREATE_STRING(a) ___CREATE_STRING(a)
#define __CLASSNAMESTRING __CREATE_STRING(MCLASSNAME)

//======================================================================
//======================================================================
//======================================================================
//
// Actual class function definition
//

// ================================================== Constructors
TEMPLATE INLINE MAT::MCLASSNAME(TYPE* new_val_ptr,int h,int w) {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(NVP=%08lx,%d,%d) this=%08lx\n",__CLASSNAMESTRING,
                  (unsigned long)new_val_ptr,h,w,(unsigned long)this);
#endif
 width=w;
 height=h;
 data=new_val_ptr;
 }

TEMPLATE INLINE MAT::MCLASSNAME() {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s() w,h,data=0 this=%08lx\n",__CLASSNAMESTRING,(unsigned long)this);
#endif
 width=0;
 height=0;
 data=0;
 }

TEMPLATE INLINE MAT::MCLASSNAME(int h,int w) {
 width=w;
 height=h;
 GET_NEW_ELTS(data,w*h);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(%d,%d) data=%08lx this=%08lx\n",__CLASSNAMESTRING,
               h,w,(unsigned long)data,(unsigned long)this);
#endif
 }

TEMPLATE INLINE MAT::MCLASSNAME(int h,int w,const TYPE* vals) {
 width=w;
 height=h;
 GET_NEW_ELTS(data,w*h);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(%d,%d,%08lx) data=%08lx this=%08lx\n",__CLASSNAMESTRING,
         h,w,(unsigned long)vals,(unsigned long)data,(unsigned long)this);
#endif
 init(vals);
 }

TEMPLATE INLINE MAT::MCLASSNAME(const MAT& m) {
 width=m.width;
 height=m.height;
 GET_NEW_ELTS(data,width*height);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(const Matrix(%d,%d,%08lx,this=%08lx)&) data=%08lx this=%08lx\n",
   __CLASSNAMESTRING,m.height,m.width,(unsigned long)m.data,(unsigned long)&m,
      (unsigned long)data,(unsigned long)this);
#endif
 init(m.data);
 }

TEMPLATE INLINE MAT::MCLASSNAME(const VECT& v) {
// JESSE GCC 7/1/98
// if(v.type==RowV) {
 if(v._type==RowV) {
   width=v.numelts;
   height=1;
   }
 else {
   width=1;
   height=v.numelts;
   }
 GET_NEW_ELTS(data,width*height);
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"%s(const Vector(%d,type=%d,%08lx,this=%08lx)&) data=%08lx this=%08lx\n",__CLASSNAMESTRING,v.numelts,(int)v._type,(unsigned long)v.data,(unsigned long)&v,
                  (unsigned long)data,(unsigned long)this);
#endif
 init(v.data);
 }

TEMPLATE INLINE MAT::~MCLASSNAME() {
#ifdef DEBUG_CTORS_AND_DTORS
 fprintf(DEBUG_SINK,"~%s() h,w=%d,%d data=%08lx this=%08lx\n",__CLASSNAMESTRING,
         width,height,(unsigned long)data,(unsigned long)this);
#endif
 width=-1;
 height=-1;
 delete data;
 data=0;
 }

//=============================================================================
//=============================================================================
//
// ** Information (where p is precision, e.g. assume zero if a(i,j) < p
//
TEMPLATE INLINE TYPE& MAT::Min(int* ip,int* jp) const
{
 CHECK_FOR_VALID_MATRIX(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   if(jp!=0)   *jp=-1;
   return (error_return_value());
   }
#endif

 // ir,jr = temp for holding location of min
 register int ir=0;
 register int jr=0;
 TYPE val=data[0];
 FOR_EACH_ELEMENT
   if(ELT<val) { ir=ROW_INDEX; jr=COL_INDEX; val=ELT; }

 // save ir,jr in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 if(jp!=0)  *jp=jr;
 return ELTK(ir,jr);
}

// return Minimum absolute value elt
TEMPLATE INLINE TYPE& MAT::AbsMin(int* ip,int* jp) const
{
 CHECK_FOR_VALID_MATRIX(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   if(jp!=0)   *jp=-1;
   return (error_return_value());
   }
#endif

 // ir,jr = temp for holding location of min
 register int ir=0;
 register int jr=0;
 TYPE val=ABS(data[0]);
 register TYPE k;
 FOR_EACH_ELEMENT {
   k=ABSELT;
   if(k<val) { ir=ROW_INDEX; jr=COL_INDEX; val=k; }
   }

 // save ir,jr in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 if(jp!=0)  *jp=jr;
 return ELTK(ir,jr);
}

//================================================== return Maximum elt
TEMPLATE INLINE TYPE& MAT::Max(int* ip,int* jp) const
{
 CHECK_FOR_VALID_MATRIX(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   if(jp!=0)   *jp=-1;
   return (error_return_value());
   }
#endif

 // ir,jr = temp for holding location of min
 register int ir=0;
 register int jr=0;
 TYPE val=data[0];
 register TYPE k;
 FOR_EACH_ELEMENT {
   k=ELT;
   if(k>val) { ir=ROW_INDEX; jr=COL_INDEX; val=k; }
   }

 // save ir,jr in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 if(jp!=0)  *jp=jr;
 return ELTK(ir,jr);
}

//================================================== return Maximum elt, absolute value
TEMPLATE INLINE TYPE& MAT::AbsMax(int* ip,int* jp) const
{
 CHECK_FOR_VALID_MATRIX(error_return_value())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 // if this is a zero size matrix, return error_zero, set i,j=0 if applicable
 if(data==0) {
   if(ip!=0)   *ip=-1;
   if(jp!=0)   *jp=-1;
   return (error_return_value());
   }
#endif

 // ir,jr = temp for holding location of min
 register int ir=0;
 register int jr=0;
 TYPE val=ABS(data[0]);
 register TYPE k;
 FOR_EACH_ELEMENT {
   k=ABSELT;
   if(k>val) { ir=ROW_INDEX; jr=COL_INDEX; val=k; }
   }

 // save ir,jr in user space if requested, then return the elt ref
 if(ip!=0)  *ip=ir;
 if(jp!=0)  *jp=jr;
 return ELTK(ir,jr);
}

//================================================== return number of elts <= p
TEMPLATE INLINE int MAT::NumZeros(const TYPER pin) const
{
 CHECK_FOR_VALID_MATRIX(-1)
 const TYPE p=ABS(pin);
 register int ct=0;
 FOR_EACH_ELEMENT {
   register TYPE k=ABSELT;
   if(k<=p) ct++;
   }
 return ct;
}

 //================================================== Predicates
 // ** Predicates (where p is precision, e.g.
 //                assume zero if a(i,j) < p

//================================================== true if this is zero
TEMPLATE INLINE int MAT::iszero(const TYPER p) const
   { return (NumZeros(p)==(width*height)); }

//======================================================================
//
// ** Element access
//
TEMPLATE INLINE TYPE& MAT::operator()(int i,int j) {
#ifdef PROTECT_VECTOR_ACCESS
#if SUBSCRIPTION_OFFSET
 if(i<0) {
#else
 if(i<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(MatrixT_Row_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(i>=(height-SUBSCRIPTION_OFFSET)) {
#else
 if(i>=height) {
#endif
   ERROR(MatrixT_Row_Bound_Too_High,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(j<0) {
#else
 if(j<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(MatrixT_Row_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(j>=(width-SUBSCRIPTION_OFFSET)) {
#else
 if(j>=width) {
#endif
   ERROR(MatrixT_Row_Bound_Too_High,error_return_value());
   }
 if((width<0)||(height<0)) {
   ERROR(MatrixT_Access_after_object_was_destroyed,error_return_value());
   }
#endif
#ifdef PROTECT_POINTER_ACCESS
 if(data==0) {
   ERROR(MatrixT_Data_Pointer_Null,error_return_value());
   }
#endif
#if SUBSCRIPTION_OFFSET
 return data[j-SUBSCRIPTION_OFFSET+(i-SUBSCRIPTION_OFFSET)*width];
#else
 return data[j+i*width];
#endif
 } // }}}}


//==================================================  operator()

TEMPLATE INLINE TYPE& MAT::operator()(int i,int j) const {
#ifdef PROTECT_VECTOR_ACCESS
#if SUBSCRIPTION_OFFSET
 if(i<0) {
#else
 if(i<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(MatrixT_Row_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(i>=(height-SUBSCRIPTION_OFFSET)) {
#else
 if(i>=height) {
#endif
   ERROR(MatrixT_Row_Bound_Too_High,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(j<0) {
#else
 if(j<SUBSCRIPTION_OFFSET) {
#endif
   ERROR(MatrixT_Row_Bound_Too_Low,error_return_value());
   }
#if SUBSCRIPTION_OFFSET
 if(j>=(width-SUBSCRIPTION_OFFSET)) {
#else
 if(j>=width) {
#endif
   ERROR(MatrixT_Row_Bound_Too_High,error_return_value());
   }
 if((width<0)||(height<0)) {
   ERROR(MatrixT_Access_after_object_was_destroyed,error_return_value());
   }
#endif
#ifdef PROTECT_POINTER_ACCESS
 if(data==0) {
   ERROR(MatrixT_Data_Pointer_Null,error_return_value());
   }
#endif
#if SUBSCRIPTION_OFFSET
 return data[j-SUBSCRIPTION_OFFSET+(i-SUBSCRIPTION_OFFSET)*width];
#else
 return data[j+i*width];
#endif
 } // }}}}

TEMPLATE INLINE VECT  MAT::row(int i) const
{
 CHECK_FOR_VALID_MATRIX(VECT())
 i-=SUBSCRIPTION_OFFSET;
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(i<0)
   ERROR(MatrixT_Attempt_To_Return_Row_With_Row_Index_Too_Low,VECT());
 if(i>=height)
   ERROR(MatrixT_Attempt_To_Return_Row_With_Row_Index_Too_High,VECT());
#endif

 TYPE* vals; GET_NEW_ELTS(vals,width);
 for(register int j=0;j<width;j++)
   vals[i]=ELTKN(i,j);
 return VECT(vals,width,RowV);
}

TEMPLATE INLINE VECT MAT::col(int j) const
{
 TYPE* vals; GET_NEW_ELTS(vals,height);
 for(register int i=0;i<height;i++)
   vals[i]=(*this)(i,j);
 return VECT(vals,height,ColV);
}

TEMPLATE INLINE VECT MAT::diag(int d) const
{
 int diag_size=(height>width?width:height);
 TYPE* vals; GET_NEW_ELTS(vals,diag_size);
 int i,j;
 if(d<0) {
   // subdiagonal
   d=-d;
   // if the requested subdiagonal starts at a row lower
   // than the height of the matrix, return a null vector
   if(d>=height)
      return VECT();
   i=SUBSCRIPTION_OFFSET+d;
   j=SUBSCRIPTION_OFFSET;
   for(register int k=d;k<height;k++) {
      vals[k-d]=(*this)(i,j);
      i++; j++;
      }
   return VECT(vals,height-d,ColV);
   }
 else {
   // if the requested subdiagonal starts at a col further left 
   // than the width of the matrix, return a null vector
   if(d>=width)
      return VECT();
   i=SUBSCRIPTION_OFFSET;
   j=SUBSCRIPTION_OFFSET+d;
   for(register int k=d;k<width;k++) {
      vals[k-d]=(*this)(i,j);
      i++; j++;
      }
   return VECT(vals,width-d,ColV);
   }
}


TEMPLATE INLINE MAT MAT::sub_matrix( int ulx, int uly, int lrx, int lry) const
{
 if(lrx<ulx) { int tmp=ulx; ulx=lrx; lrx=tmp; }
 if(lry<uly) { int tmp=uly; uly=lry; lry=tmp; }
 int ret_width =lrx-ulx;
 int ret_height=lry-uly;
 if(ulx<SUBSCRIPTION_OFFSET) ulx=SUBSCRIPTION_OFFSET;
   else ulx-=SUBSCRIPTION_OFFSET;
 if(uly<SUBSCRIPTION_OFFSET) uly=SUBSCRIPTION_OFFSET;
   else uly-=SUBSCRIPTION_OFFSET;
 if((lrx-SUBSCRIPTION_OFFSET)>=width) lrx=width-SUBSCRIPTION_OFFSET;
   else lrx-=SUBSCRIPTION_OFFSET;
 if((lry-SUBSCRIPTION_OFFSET)>=height) lry=width-SUBSCRIPTION_OFFSET;
   else lry-=SUBSCRIPTION_OFFSET;
 TYPE* vals; GET_NEW_ELTS(vals,ret_width*ret_height);
 int val_index=0;
 for(int j=uly;j<lry;j++)
   for(int i=ulx;i<lrx;i++) {
      vals[val_index]=(*this)(i,j);
      val_index++;
      }
 return MAT(vals,ret_height,ret_width);
}

//================================================== General Matrix Operations

//==================================================
// init from array
TEMPLATE INLINE MAT& MAT::init(const TYPE* val) {
 CHECK_FOR_VALID_MATRIX((*this))
#ifdef PROTECT_POINTER_ACCESS
 if(val==NULL)
   ERROR(MatrixT_Null_Pointer_To_Init,(*this));
#endif
 register int k=0;
 FOR_EACH_ELEMENT
   ELT = val[k++];
 return *this;
 }

//==================================================
// set all elts = val
TEMPLATE INLINE MAT& MAT::fill(const TYPER val) {
 CHECK_FOR_VALID_MATRIX((*this))
 FOR_EACH_ELEMENT
   ELT=val;
 return (*this);
 }

//==================================================
// take absolute value of each elt
TEMPLATE INLINE MAT& MAT::abs() {
 CHECK_FOR_VALID_MATRIX((*this))
 FOR_EACH_ELEMENT
   ELT=ABSELT;
 return (*this);
 }

//==================================================
// resize matrix to i x j
TEMPLATE INLINE MAT& MAT::reshape(int i,int j) {
 CHECK_FOR_VALID_MATRIX((*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if((width*height)!=(i*j))
   ERROR(MatrixT_Attempt_To_Reshape_To_Invalid_Sizes,(*this));
#endif
 height=i;
 width=j;
 return (*this);
 }


TEMPLATE INLINE MAT& MAT::replace_column(int j,const VECT& v) {
 CHECK_FOR_VALID_MATRIX((*this))
 CHECK_V_FOR_VALID_VECTOR(v,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(j<SUBSCRIPTION_OFFSET)
   ERROR(MatrixT_Replace_Column_Attempted_With_Column_Value_Too_Low,(*this))
 else
   if(j>=(width-SUBSCRIPTION_OFFSET))
      ERROR(MatrixT_Replace_Column_Attempted_With_Column_Value_Too_High,(*this))
#endif

 int row_max=0;
 // next we check to see if the vector is larger or smaller
 // than the matrix height
 if(v.numelts<height)
   row_max=v.numelts;
 else
   row_max=height;

 register int k=SUBSCRIPTION_OFFSET;
 row_max+=SUBSCRIPTION_OFFSET;
 register int COL_INDEX=j;
 for(register int ROW_INDEX=SUBSCRIPTION_OFFSET;ROW_INDEX<row_max;ROW_INDEX++)
   ELT=v(k++);

 return (*this);
 }

TEMPLATE INLINE MAT& MAT::replace_row(int i,const VECT& v) {
 CHECK_FOR_VALID_MATRIX((*this))
 CHECK_V_FOR_VALID_VECTOR(v,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(i<SUBSCRIPTION_OFFSET)
   ERROR(MatrixT_Replace_Column_Attempted_With_Column_Value_Too_Low,(*this))
 else
   if(i>=(height-SUBSCRIPTION_OFFSET))
      ERROR(MatrixT_Replace_Column_Attempted_With_Column_Value_Too_High,(*this))
#endif

 int col_max=0;
 // next we check to see if the vector is larger or smaller
 // than the matrix width
 if(v.numelts<width)
   col_max=v.numelts;
 else
   col_max=width;

 register int k=SUBSCRIPTION_OFFSET;
 col_max+=SUBSCRIPTION_OFFSET;
 register int ROW_INDEX=i;
 for(register int COL_INDEX=SUBSCRIPTION_OFFSET;COL_INDEX<col_max;COL_INDEX++)
   ELT=v(k++);

 return (*this);
 }

//==================================================
// true if this(i,j) = a(i,j)
TEMPLATE INLINE int MAT::operator ==(const MAT&  m) const
{
 if(m.width!=width) return 0;
 if(m.height!=height) return 0;
 FOR_EACH_ELEMENT
   if(ELT != REFELT(m)) return 0;

 return 1;
}

TEMPLATE INLINE int MAT::operator !=(const MAT&  m) const
   { return (!((*this)==m)); }

// true if this(i,j) >= a(i,j)
TEMPLATE INLINE int MAT::operator >=(const MAT&  m) const
{
 if((m.width*m.height)<(width*height)) return 0;
 FOR_EACH_ELEMENT
   if(ELT < REFELT(m)) return 0;

 return 1;
}

// true if this(i,j) <= a(i,j)
TEMPLATE INLINE int MAT::operator <=(const MAT&  m) const
{
 if((m.width*m.height)>(width*height)) return 0;
 FOR_EACH_ELEMENT
   if(ELT > REFELT(m)) return 0;

 return 1;
}

// true if this(i,j) > a(i,j)
TEMPLATE INLINE int MAT::operator >(const MAT&  m) const
{
 if((m.width*m.height)<=(width*height)) return 0;
 FOR_EACH_ELEMENT
   if(ELT <= REFELT(m)) return 0;

 return 1;
}

// true if this(i,j) < a(i,j)
TEMPLATE INLINE int MAT::operator <(const MAT&  m) const
{
 if((m.width*m.height)>=(width*height)) return 0;
 FOR_EACH_ELEMENT
   if(ELT >= REFELT(m)) return 0;
 return 1;
}

 //================================================== Assignment Operators
// this <- m
TEMPLATE INLINE MAT& MAT::operator =(const MAT&  m)
{
 if(data!=0) { // then this is not a null matrix so we check to make sure the
               // sizes agree, then we use the same space as the other one.
   if((m.width*m.height)<(height*width)) {
      width=m.width;
      height=m.height;
      }
   else {
      delete data;
      width=m.width;
      height=m.height;
      GET_NEW_ELTS(data,width*height);
      }
   }
 else {
   width=m.width;
   height=m.height;
   GET_NEW_ELTS(data,width*height);
   }

 init(m.data);
 return (*this);
}

TEMPLATE INLINE MAT::operator double() const {
 CHECK_FOR_VALID_MATRIX(0.0)
 return (double)data[0];
 }

TEMPLATE INLINE MAT::operator long() const {
 CHECK_FOR_VALID_MATRIX(0)
 return (long)data[0];
 }

TEMPLATE INLINE MAT& MAT::operator +=(const MAT& m) {
 CHECK_FOR_VALID_MATRIX((*this))
 CHECK_M_FOR_VALID_MATRIX(m,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if((m.width!=width)||(m.height!=height))
   ERROR(MatrixT_Attempted_Plus_Equals_With_Different_Dimensions,(*this))
#endif
 FOR_EACH_ELEMENT
   ELT+=REFELT(m);
 return (*this);
}

TEMPLATE INLINE MAT& MAT::operator -=(const MAT& m) {
 CHECK_FOR_VALID_MATRIX((*this))
 CHECK_M_FOR_VALID_MATRIX(m,(*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if((m.width!=width)||(m.height!=height))
   ERROR(MatrixT_Attempted_Minus_Equals_With_Different_Dimensions,(*this))
#endif
 FOR_EACH_ELEMENT
   ELT-=REFELT(m);
 return (*this);
}


TEMPLATE INLINE MAT& MAT::operator /=(const TYPER s) {
 CHECK_FOR_VALID_MATRIX((*this))
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(s==(TYPE)0)
   ERROR(MatrixT_Attempt_To_Divide_Equals_By_Zero,(*this))
#endif
 FOR_EACH_ELEMENT
   ELT/=s;
 return (*this);
 }

TEMPLATE INLINE MAT& MAT::operator *=(const TYPER s) {
 CHECK_FOR_VALID_MATRIX((*this))
 FOR_EACH_ELEMENT
   ELT*=s;
 return (*this);
 }

//================================================== Matrix Operators
// matrix inverse
TEMPLATE INLINE MAT MAT::operator !() const {
 printf("MATRIX INVERT OPERATOR (!) NOT YET IMPLEMENTED.\n");
 return *this;
}

TEMPLATE INLINE MAT MAT::inv() const {
 printf("MATRIX INVERT NOT YET IMPLEMENTED.\n");
 return (*this);
}

TEMPLATE INLINE MAT& MAT::inv_self() {
 printf("MATRIX INVERT SELF NOT YET IMPLEMENTED.\n");
 return *this;
 }

// matrix transpose
TEMPLATE INLINE MAT MAT::operator ~() const {
 CHECK_FOR_VALID_MATRIX(MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 for(register int j=0;j<width;j++)
   for(register int i=0;i<height;i++)
      { vals[k]=(*this)(i,j); k++; }
 return MAT(vals,width,height);
}

TEMPLATE INLINE MAT MAT::transpose() const {
 CHECK_FOR_VALID_MATRIX(MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 for(register int j=0;j<width;j++)
   for(register int i=0;i<height;i++)
      { vals[k]=(*this)(i,j); k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT& MAT::transpose_self() {
 CHECK_FOR_VALID_MATRIX((*this))
 TYPE* vals=new TYPE[width*height];
 register int k=0;
 for(register int j=0;j<width;j++)
   for(register int i=0;i<height;i++)
      { vals[k]=(*this)(i,j); k++; }
 int tmp=width;
 width=height;
 height=tmp;
 init(vals);
 delete[] vals;
 return (*this);
}



TEMPLATE INLINE MAT MAT::operator +(const MAT&  m) const {
 CHECK_FOR_VALID_MATRIX(MAT())
 CHECK_M_FOR_VALID_MATRIX(m,MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 FOR_EACH_ELEMENT
   { vals[k]=ELT+REFELT(m); k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT MAT::operator -() const {
 CHECK_FOR_VALID_MATRIX(MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 FOR_EACH_ELEMENT
   { vals[k]=-ELT; k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT MAT::operator -(const MAT&  m) const {
 CHECK_FOR_VALID_MATRIX(MAT())
 CHECK_M_FOR_VALID_MATRIX(m,MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 FOR_EACH_ELEMENT
   { vals[k]=ELT-REFELT(m); k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT MAT::operator /(const TYPER  s) const {
 CHECK_FOR_VALID_MATRIX(MAT())
#ifndef DONT_PROTECT_INVALID_OPERATIONS
 if(s==(TYPE)0)
   ERROR(MatrixT_Attempted_Scalar_Division_By_Zero,MAT())
#endif
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 FOR_EACH_ELEMENT
   { vals[k]=ELT/s; k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT MAT::operator *(const TYPER  s) const {
// Jesse Reichler fixed what was a division 12/18/96
 CHECK_FOR_VALID_MATRIX(MAT())
 TYPE* vals;
 GET_NEW_ELTS(vals,width*height);
 register int k=0;
 FOR_EACH_ELEMENT
//   { vals[k]=ELT/s; k++; }
   { vals[k]=ELT*s; k++; }
 return MAT(vals,width,height);
 }

TEMPLATE INLINE MAT MAT::operator *(const VECT&  v) const {
 CHECK_FOR_VALID_MATRIX(MAT())
 CHECK_V_FOR_VALID_VECTOR(v,MAT())
 if(v._type==RowV) {
   MAT tmp(1,v.numelts);
	tmp.init(v.data);
   return MAT((*this)*tmp);
   }
 else {
   MAT tmp(v.numelts,1);
	tmp.init(v.data);
   return MAT((*this)*tmp);
   }
 }

TEMPLATE INLINE MAT MAT::operator *(const MAT&  m) const {
 CHECK_FOR_VALID_MATRIX(MAT())
 CHECK_M_FOR_VALID_MATRIX(m,MAT())
 MAT tret(height,m.width);
 register int col=SUBSCRIPTION_OFFSET;
 for(;col<tret.width+SUBSCRIPTION_OFFSET;col++)
   for(register int row=SUBSCRIPTION_OFFSET;row<tret.height+SUBSCRIPTION_OFFSET;row++) {
      TYPE val=0.0;
      register int mult=SUBSCRIPTION_OFFSET;
      for(;mult<width+SUBSCRIPTION_OFFSET;mult++)
         val+=ELTK(row,mult)*REFELTK(m,mult,col);
      REFELTK(tret,row,col)=val;
      }
 return MAT(tret);
 }

// JESSE GCC 7/1/98
TEMPLATE INLINE MAT operator *(const TYPER s,const MAT& m) { return m*s; }
/*
TEMPLATE INLINE MAT operator *(const VECT& v,const MAT& m) {
 return MAT(m*v);
 }
 */

 //================================================== Matrix IO
// print an ascii version of the matrix
TEMPLATE INLINE void MAT::print(FILE* file) const {
 fprintf(file, "%d %d\n[",height, width);
 int k=0;
 FOR_EACH_ELEMENT {
   fprintf(file, " %g", (double)ELT);
   if((++k)==width) {
      fprintf(file, ";\n " );
      k=0;
      }
   }
 fprintf(file, "]\n" );
}

// read an ascii version of the matrix
TEMPLATE INLINE int MAT::read(FILE* file)
{
 if(data!=0)
   FREE_ELT_SPACE(data);
 
 fscanf(file, "%d %d\n[ ", &height, &width);

 GET_NEW_ELTS(data,width*height);
 for (int i=0; i<height; i++) {
   double d;
   for (int j=0; j<width; j++)
   {
      fscanf(file, "%lf", &d);
      (*this)(i,j)=(TYPE)d;
      }
 fscanf(file, ";\n");
 }
 fscanf(file, "]\n");

 return 0;
}


// save to f in format fmt
TEMPLATE INLINE int MAT::save(FILE* f,const char* fmt) const
{
 if(strcmp(fmt,"img")==0) {
   fprintf(f,"width=%d\n",width);
   fprintf(f,"height=%d\n",height);
   if(fwrite((char*)data,sizeof(TYPE),width*height,f)==0) {
      return -1;
      }
   } else
 if(strcmp(fmt,"data")==0) {
   if(fwrite((char*)data,sizeof(TYPE),width*height,f)==0) {
      return -1;
      }
   } else
 if(strcmp(fmt,"raw")==0) {
   fprintf(f,"P2\n#%s\n%d %d\n%d\n","MTX",width,height,255);
   int i,j,k;
   for(i=0;i<height;i++)
      for(j=0;j<width;j+=25) {
         int max;
         if(j+25<width) max=25; else max=width-j;
         for(k=0;k<max;k++)
            fprintf(f," %3ld",(long)(*this)(i,j+k));
         fprintf(f,"\n");
         }
   } else
 if(strcmp(fmt,"matlab")==0) {
   long mopt=1000;
   long mrows=height;
   long mcols=width;
   long imagf=0;
   long namelen=strlen("MTX")+1;
   if(fwrite((char*)&mopt,sizeof(long),1,f)==0) return -1;
   if(fwrite((char*)&mrows,sizeof(long),1,f)==0) return -1;
   if(fwrite((char*)&mcols,sizeof(long),1,f)==0) return -1;
   if(fwrite((char*)&imagf,sizeof(long),1,f)==0) return -1;
   if(fwrite((char*)&namelen,sizeof(long),1,f)==0) return -1;
   if(fwrite((char*)"MTX",strlen("MTX")+1,sizeof(char),f)==0) return -1;
   double val;
   int i,j;
   for(j=0;j<width;j++)
      for(i=height-1;i>=0;i--) {
         val=(double)(*this)(i,j);
         if(fwrite((char*)&val,sizeof(double),1,f)==0) return -1;
         }
   }
 return 0;
}

// create fname and save, use format fmt
TEMPLATE INLINE int MAT::save(const char* fname,const char* fmt) const
{
 FILE* f;
 if((f=fopen(fname,"w"))==0)
   return -1;
 int ret;
 ret=save(f,fmt);
 fclose(f);
 return ret;
}

// load from f, expect format fmt
TEMPLATE INLINE int MAT::load(FILE* f,const char* fmt)
{
 if(strcmp(fmt,"img")==0) {
   delete data;
   if(fscanf(f,"width=%d\n",&width)!=1)   return -1;
   if(fscanf(f,"height=%d\n",&height)!=1) return -1;
   GET_NEW_ELTS(data,width*height);
   if(fread((char*)data,sizeof(TYPE),width*height,f)==0) {
      return -1;
      }
   }
 return 0;
}

// open fname and load, expect format fmt
TEMPLATE INLINE int MAT::load(const char* fname,const char* fmt) {
 FILE* f;
 if((f=fopen(fname,"r"))==NULL) return -1;
 return load(f,fmt);
}


#undef INLINE
#undef ROW_INDEX
#undef COL_INDEX 
#undef FOR_EACH_ELEMENT
#undef FOR_EACH_ROW_ELEMENT
#undef FOR_EACH_COL_ELEMENT
#undef REFELT
#undef REFELTK
#undef REFELTKN
#undef ELT
#undef ELTK
#undef ELTKN
#undef ABS
#undef ABSELT
#undef GET_NEW_ELTS
#undef FREE_ELT_SPACE
#undef CHECK_FOR_VALID_VECTOR
#undef CHECK_V_FOR_VALID_VECTOR
#undef CHECK_FOR_VALID_MATRIX
#undef CHECK_M_FOR_VALID_MATRIX
#undef ___CREATE_STRING
#undef __CREATE_STRING
#undef __CLASSNAMESTRING

#ifndef __DONT_DEFINE_VECTOR_TEMPLATE
#undef TYPE
#undef TYPER
#undef VCLASSNAME
#undef MCLASSNAME
#endif
#undef TEMPLATE
#undef MAT
#undef VECT

#endif
