// JRutils.h
// This file defines miscelaneous utility functions.

// forward declarations
class JRAliasmap;
class JRenum;

// start file
#ifndef INC_JRutilsh
#define INC_JRutilsh

#include <fstream>
#include <list>
#include <map>
#include <string>



// we need a char array to hold string conversions of numbers.  actually the size
//  should be at least 2*max_number_string + ~10 to hold a signed range.
#define NUMBERSTR_MAXSIZE	255

// internal kludge constants
#define PRIORLINE_LEFT		-9999
#define NEXTLINE_RIGHT		 9999

// for nice printing
#define MAX_DISPLAYLINELENGTH	200

// These constants describe how erros are highlighted.
//  on a linux console and other vts that support color escape sequences,
//  we can do a nice color identification, in which case the
//  bool INSERT_MULTILINE_MARKS should be true, (which restores
//  the default between multi-line errors.
//  otherwise false looks better if you are using standard characters to
//  indicate location of errors.

// for pure text term
//#define INSERT_MULTILINE_MARKS	false
//#define HIGHLIGHT_LEFT		">>"
//#define HIGHLIGHT_RIGHT		"<<"

// for linux and others with color
#define INSERT_MULTILINE_MARKS		true
#define HIGHLIGHT_LEFT			"\033[;33;1m"
#define HIGHLIGHT_RIGHT			"\033[m"






//Not implemented in gcc yet (2.8.1)
//using namespace std;
//namespace JR{





//--------------------------------------------------------------------------------
//---- JRAliasmap ----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: A class for managing string aliases

class JRAliasmap
{
private:
  map<string,string> csaliasmap;				// case sensitive map of keywords and aliases
  map<string,string> cisaliasmap;				// caseinsensitive map of keywords and aliases
public:
  string alias(string wordstr);						// return what wordstr is an alias of
  void addalias(string alias,string keyword,bool casesensitive=true);	// add a new alias to the list of aliases
};



//--------------------------------------------------------------------------------
//---- JRenum --------------------------------------------------------------------
//--------------------------------------------------------------------------------
//: A class for managing enumeration lists

class JRenum
{
private:
  map<string,int> enumap;
public:
  JRenum() {;};
  ~JRenum() {;};
  JRenum(string enumlist) {;};
public:
  void add(string enumword,int index) {;};
  int index(string enumword,bool throwerr=true) {return 0;};
  bool isenum(string enumword) {return true;};
  string lookup(int index,bool throwerr=true) {return "enum0";};
  string range() {return "enum1,enum2";};
  static int strindex(string enumword,string enumlist,bool throwerr=true) {return 0;};
  static bool isenumstr(string enumword,string enumlist) {return true;};
  static string strlookup(int index,bool throwerr=true) {return "enum0";};
  static string strrange(string enumstr) {return "enum1,enum2";};
};









//--------------------------------------------------------------------------------
//---- JRSearchFileOpen() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: Find and open the file stream for input, searching through directory list to find.
// Subtlety: if we do find it someplace unusual, change filename to the actual location
//  what we do is we search our list of directories, checking for first the exact filename passed,
//  and then the filename+extension.

fstream *JRSearchFileOpen(string filename, list<string> *dirs, string &outfilename, ios::openmode mode=ios::in, const string ext="");





//--------------------------------------------------------------------------------
//---- JRShowStreamPos() ---------------------------------------------------------
//--------------------------------------------------------------------------------
//: Display a specified line from a file and highlight selected section.
// Finds the closest line and character in the file and display the line,
//  line #, and put carrots under the error location.
// This is mainly used as an error-reporting helper function:

void JRShowStreamPos(ostream *out, string sname, istream *sisp, long sline, long slinepos, string ename="", istream *eisp=NULL, long eline=0, long elinepos=0);


//--------------------------------------------------------------------------------
//---- JRShowStreamLine() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: Helper for ShowStreamPos.
// Show the line from the stream, and carrot positions if specified

void JRShowStreamLine(ostream *out, istream *isp, long sline, long slinepos=0, long elinepos=0);



//--------------------------------------------------------------------------------
//---- JRGetStreamLine() ---------------------------------------------------------
//--------------------------------------------------------------------------------
//: Helper for ShowStreamLine.
//  Read the line specified, return fileposition.

string JRGetStreamLine(ostream *out, istream *isp, long sline);



//--------------------------------------------------------------------------------
//---- JRstringcharreplace() -----------------------------------------------------
//--------------------------------------------------------------------------------
//: Replace all occurences of fchar with rchar.

void JRstringcharreplace(string &instring, char fchar, char rchar);




//--------------------------------------------------------------------------------
//---- JRsplicearound() ----------------------------------------------------------
//--------------------------------------------------------------------------------
//: Insert two strings inside another, to the left and right of two positions

void JRmysplicearound(string &instring,long leftpos,string leftinsert,long rightpos=-1,string rightinsert="");




//--------------------------------------------------------------------------------
//---- JRlowerstring() -----------------------------------------------------------
//--------------------------------------------------------------------------------
//: return a lower case version of the string

string JRlowerstring(string instring);




//--------------------------------------------------------------------------------
//---- JRclcmp() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: a commandline flag helper function
// return true if the flag is in the word.    AND remove the matching flag

bool JRclcmp(char *comword,char *shortstr,char *longstr);




//--------------------------------------------------------------------------------
//---- JRstripword()
//--------------------------------------------------------------------------------
//: strip off and return the leading word (before space or \0, or comma)

string JRstripword(string &instring);



//--------------------------------------------------------------------------------
//---- JRstriptoken()
//--------------------------------------------------------------------------------
//: strip off and return the next token

string JRstriptoken(string &instring);


//--------------------------------------------------------------------------------
//---- JRcleanstring()
//--------------------------------------------------------------------------------
//: strip off and return any leading or trailing spaces

void JRcleanstring(string &instring);



//--------------------------------------------------------------------------------
//---- JRstripline()
//--------------------------------------------------------------------------------
//: strip off and return the leading line (before \n or \0)

string JRstripline(string &instring);




//--------------------------------------------------------------------------------
//---- JRstripleader() -----------------------------------------------------------
//--------------------------------------------------------------------------------
//: strip off and return any leading trail (before first period) or return ""

string JRstripleader(string &instring);



//--------------------------------------------------------------------------------
//---- JRstriptail() -------------------------------------------------------------
//--------------------------------------------------------------------------------
//: strip off and return any trail (before first period) or return instring
// this is just like stripleader except that we return the tail not leader

// first version considers a string with no period to have no tail
string JRstriptail(string &instring);
// the second version considers a string with no period to be a full tail
string JRstriptail2(string &instring);



//--------------------------------------------------------------------------------
//---- JRinvalidcomargs() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: return an error if there are any unaccounted for (non-blank) arguments on the commandline
// if string pointer is passed, it will be filled with a string list of the unknown args

bool JRinvalidcomargs(int argc, char **argv, bool complain=true, string *unknownargstr=NULL);



//--------------------------------------------------------------------------------
//---- JRcompresscomargs() -------------------------------------------------------
//--------------------------------------------------------------------------------
//: any blank command line args are removed.  this function is used when other funcs
// remove (zero-out) commandline args as they consume them.
// RETURNS the new number of commandline arguments

void JRcompresscomargs(int &argc, char **argv);


//--------------------------------------------------------------------------------
//---- JRstripcomarg() -----------------------------------------------------------
//--------------------------------------------------------------------------------
//: remove the leading non-blank commandline argument and return it, setting new argc

string JRstripcomarg(int &argc,char **argv);



//--------------------------------------------------------------------------------
//---- JRremquotes() -------------------------------------------------------------
//--------------------------------------------------------------------------------
//: remove enclosing quotes from string

void JRremquotes(string &instring);



//--------------------------------------------------------------------------------
//---- JRsimiliarstr() -----------------------------------------------------------
//--------------------------------------------------------------------------------
//: return true if the strings are "similiar" enough (used for warnings, hints, etc.)

bool JRsimiliarstr(string str1,string str2);



//--------------------------------------------------------------------------------
//---- JRescaperep() -------------------------------------------------------------
//--------------------------------------------------------------------------------
//: return a string with any escaped symbols (\n,\t, etc) replace with their
// ascii equivelants

string JRescaperep(string instring);




//--------------------------------------------------------------------------------
//---- JRfindandreplace() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: find and replace substrings with characters

void JRfindandreplace(string &instring,string findstr,char repc);


//--------------------------------------------------------------------------------
//---- JRfindandreplace() --------------------------------------------------------
//--------------------------------------------------------------------------------
//: find and replace substrings with substrings (beware infinite recursions)

void JRfindandreplace(string &instring,string findstr,string repstr);



//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: convert argument to a string (should be overloaded for dif. arg types)
string JRtostr(double val);


//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: convert argument to a string (should be overloaded for dif. arg types)
string JRtostr(int val);


//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: convert argument to a string (should be overloaded for dif. arg types)
string JRtostr(long val);



//--------------------------------------------------------------------------------
//---- JRtostr() -----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: convert argument to a string (should be overloaded for dif. arg types)
string JRtostr(char *str);



//--------------------------------------------------------------------------------
//---- JRdtostr() ----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: return a string which is a "%lg" sprintf formated printing of a long double
// if formatstr is non-blank, it will be used in place of "%lg"
string JRdtostr(double val,string formatstr="");



//--------------------------------------------------------------------------------
//---- JRstradr()
//--------------------------------------------------------------------------------
//: return c_str() or NULL if blank
const char *JRstradr(string *instring);



//--------------------------------------------------------------------------------
//---- JRrand01() ----------------------------------------------------------------
//--------------------------------------------------------------------------------
//: return a (double) uniform random number [0,1]
double JRrand01();








//--------------------------------------------------------------------------------
//---- JRcleancharp()
//--------------------------------------------------------------------------------
//: remove anything after comments

void JRcleancharp(char *instring, string commentprefix);


//--------------------------------------------------------------------------------
//---- JRstriptok()
//--------------------------------------------------------------------------------
//: strip out the next token (separated by commas or spaces)
void JRstriptok(char *instring,char *outword);


//--------------------------------------------------------------------------------
//---- JRfprintf()
//--------------------------------------------------------------------------------
//: sprintf then send to output stream
void JRfprintf(ostream *ostreamp,const char *str, ...);


//--------------------------------------------------------------------------------
//---- JRstrprintf()
//--------------------------------------------------------------------------------
//: sprintf then convert to string
string JRstrprintf(const char *str, ...);




//--------------------------------------------------------------------------------
//---- JRsgn()
//--------------------------------------------------------------------------------
//: ridiculousness
float JRsgn(float val);







// }
// Namespace close-brace
#endif



